require 'shoulda/matchers/integrations/libraries/action_controller'
require 'shoulda/matchers/integrations/libraries/active_model'
require 'shoulda/matchers/integrations/libraries/active_record'
require 'shoulda/matchers/integrations/libraries/missing_library'
require 'shoulda/matchers/integrations/libraries/rails'
require 'shoulda/matchers/integrations/libraries/routing'

module Shoulda
  module Matchers
    module Integrations
      # @private
      module Libraries
      end
    end
  end
end
