require 'shoulda/matchers/integrations/test_frameworks/active_support_test_case'
require 'shoulda/matchers/integrations/test_frameworks/minitest_4'
require 'shoulda/matchers/integrations/test_frameworks/minitest_5'
require 'shoulda/matchers/integrations/test_frameworks/missing_test_framework'
require 'shoulda/matchers/integrations/test_frameworks/rspec'
require 'shoulda/matchers/integrations/test_frameworks/test_unit'

module Shoulda
  module Matchers
    module Integrations
      # @private
      module TestFrameworks
      end
    end
  end
end
