require 'shoulda/matchers/independent/delegate_method_matcher'
require 'shoulda/matchers/independent/delegate_method_matcher/target_not_defined_error'

module Shoulda
  module Matchers
    # This module provides matchers that are used to test behavior outside of
    # Rails-specific classes.
    module Independent
    end
  end
end
